/* This file is part of P^nMPI.
 *
 * Copyright (c)
 *  2008-2019 Lawrence Livermore National Laboratories, United States of America
 *  2011-2016 ZIH, Technische Universitaet Dresden, Federal Republic of Germany
 *  2013-2019 RWTH Aachen University, Federal Republic of Germany
 *
 *
 * P^nMPI is free software; you can redistribute it and/or modify it under the
 * terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation version 2.1 dated February 1999.
 *
 * P^nMPI is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with P^nMPI; if not, write to the
 *
 *   Free Software Foundation, Inc.
 *   51 Franklin St, Fifth Floor
 *   Boston, MA 02110, USA
 *
 *
 * Written by Martin Schulz, schulzm@llnl.gov.
 *
 * LLNL-CODE-402774
 */

#ifndef PNMPI_PRIVATE_RETURN_ADDRESS_H
#define PNMPI_PRIVATE_RETURN_ADDRESS_H


#include <stddef.h>

#include <pnmpi/private/attributes.h>
#include <pnmpi/private/function_address.h>
#include <pnmpi/private/tls.h>


/* The following macro will be defined, if CMake has successfully checked the
 * compiler has the builtin function __builtin_return_address().
 *
 * NOTE: Usually cmakefines should be placed in a separate header file. However,
 *       as this file is private and the macro will be used in this file and the
 *       related service function only, it'll be placed here.
 */
#define HAVE_BUILTIN_RETURN_ADDRESS


extern pnmpi_compiler_tls_keyword void *pnmpi_return_address;


/** \brief Store the return address of the application's MPI call.
 *
 * \note A macro needs to be used instead of a static function, otherwise the
 *  return address would be pointing to the calling PnMPI function.
 *
 * \note If the return address has already been set, a new value will NOT be
 *  stored. This is required for the Fortran wrapper redirections, as these call
 *  the C wrappers internal. Remember to reset the return address before exiting
 *  the wrapper.
 *
 * \note If the compiler doesn't support a builtin function to get the return
 *  address, this macro will be empty and no commands will be executed.
 *
 *
 * \private
 */
#if defined(DOXYGEN) || !defined(HAVE_BUILTIN_RETURN_ADDRESS)
#define pnmpi_set_return_address()
#else
#define pnmpi_set_return_address()  \
  if (pnmpi_return_address == NULL) \
    pnmpi_return_address = __builtin_return_address(0);
#endif


/** \brief Get the return address.
 *
 *
 * \return The return address of the function call currently wrapped.
 *
 *
 * \private
 */
PNMPI_UNUSED
static void *pnmpi_get_return_address(void)
{
  return pnmpi_return_address;
}


/** \brief Reset the return address.
 *
 *
 * \private
 */
PNMPI_UNUSED
static void pnmpi_reset_return_address(void)
{
  pnmpi_return_address = NULL;
}


#endif
